from creationism.registration.factory import RegistrantFactory
from creationism.utils import first_lowered

DEFAULT_MODE = "default"


class Mode(RegistrantFactory):
    STATIC = True
    AUTO = True
    CONVERT_NAME = first_lowered


class DefaultMode(Mode):
    ...



import warnings
from abc import ABC
from typing import Tuple, Callable
from creationism.registration.warnings import DuplicateRegistrantNameWarning
from functools import wraps

class Registrar(ABC):
    """Registrable class that can be used to register subclasses.

    Class Attributes:
        AUTO (bool): when enabled a subclass will be automatically registered with its class name
        REPLACE (bool): ...
        CONVERT_NAME (Callable): function that converts the class name

    Warnings:
        DuplicateRegistrantNameWarning:

    """

    AUTO = False
    REPLACE = True
    CONVERT_NAME = None
    RECURSIVE = False

    _REGISTER = None

    @property
    def name(self):
        return self.__class__.class_name()

    @classmethod
    def names(cls):
        return tuple(cls._REGISTER)

    @classmethod
    def registered(cls, name):
        return name in cls._REGISTER

    @classmethod
    def __init_subclass__(cls):
        if cls.AUTO:
            cls._register(
                bases=cls.__bases__,
                registrant_names=(cls.class_name(),),
                replace=cls.REPLACE,
                recursive=cls.RECURSIVE
            )

    @classmethod
    def class_name(cls):
        if cls.CONVERT_NAME is None:
            return cls.__name__
        return cls.CONVERT_NAME(cls.__name__)

    @classmethod
    def get_registrant(cls, registrant_name):
        if isinstance(registrant_name, type) and registrant_name in cls._REGISTER.values():
            return registrant_name
        return cls._REGISTER[registrant_name]

    @classmethod
    def register(cls, registrant_names: Tuple, replace=False, recursive=False):
        @wraps(cls.register)
        def decorator(registrant: Registrar):
            registrant._register(
                registrant.__bases__,
                registrant_names=registrant_names,
                replace=replace,
                recursive=recursive
            )
            return registrant

        return decorator

    @classmethod
    def register_func(cls, registrant_names: Tuple, replace=False):
        @wraps(cls.register_func)
        def decorator(registrant: Callable):
            cls._initialize_REGISTER()
            cls._add_to_register(registrant, registrant_names, replace)
            return registrant
        return decorator

    @classmethod
    def _register(cls, bases: Tuple, registrant_names: Tuple[str], replace: bool, recursive: bool):
        for base in bases:
            if issubclass(base, Registrar) and base is not Registrar:
                cls._add_to_base_registers(base, registrant_names, replace)
                if recursive:
                    cls._register(
                        bases=base.__bases__,
                        registrant_names=registrant_names,
                        replace=replace,
                    )
                cls._deinitialize_REGISTER()

    @classmethod
    def _add_to_base_registers(
        cls, base: "Registrar", registrant_names: Tuple, replace: bool
    ):
        if Registrar in base.__bases__:
            return
        base._initialize_REGISTER()
        base._add_to_register(
            subclass=cls, registrant_names=registrant_names, replace=replace
        )

    @classmethod
    def _initialize_REGISTER(cls):
        if cls._REGISTER is None:
            cls._REGISTER = {}

    @classmethod
    def _deinitialize_REGISTER(cls):
        cls._REGISTER = None

    @classmethod
    def _add_to_register(
        cls, subclass: "Registrar", registrant_names: Tuple, replace: bool
    ):
        for name in registrant_names:
            if name in cls._REGISTER and not cls.REPLACE:
                warnings.warn(
                    DuplicateRegistrantNameWarning(cls, name, subclass, cls._REGISTER)
                )
                continue

            if name is not None:
                cls._REGISTER[name] = subclass



from abc import abstractstaticmethod
from typing import List

from creationism.extension import Extension
from creationism.utils import open_json, open_yaml
import os
from pathlib import Path
SEARCH_PATHS = ['']

class ConfigurationFileExtension(Extension):
    @abstractstaticmethod
    def open(file_path):
        """[summary]

        Args:
            file_path ([type]): [description]

        Returns:
            [type]: [description]
        """

    @staticmethod
    def is_configuration_extension_path(path):
        return path.endswith(ConfigurationFileExtension.names())


@ConfigurationFileExtension.register((".yml", ".yaml"))
class YamlConfigurationFileExtension(ConfigurationFileExtension):
    @staticmethod
    def open(file_path):
        return open_yaml(file_path)


@ConfigurationFileExtension.register((".json",))
class JsonConfigurationFileExtension(ConfigurationFileExtension):
    @staticmethod
    def open(file_path):
        return open_json(file_path)


def open_config(config_path: Path, search_paths: List):
    for search_path in search_paths:
        search_config_path = os.path.join(search_path, config_path)
        try:
            return ConfigurationFileExtension.create(Path(config_path).suffix).open(search_config_path)
        except FileNotFoundError:
            pass
    else:
        raise FileNotFoundError(f"config not found: {config_path}")


import logging
from typing import Optional, List
import os
import pathlib
from datetime import datetime
from typing import Optional

_logger_names = []
_log_path = None


def _get_log_path() -> str:
    global _log_path
    if _log_path is None:
        raise ValueError("log_path not set")
    return _log_path


def reset_logging(log_path: str, time_stamped=True) -> Optional[str]:
    global _logger_names
    for name in _logger_names:
        logger = logging.getLogger(name)
        if isinstance(logger, logging.Logger):
            logger.setLevel(logging.NOTSET)
            logger.propagate = True
            logger.disabled = False
            logger.filters.clear()
            handlers = logger.handlers.copy()
            for handler in handlers:
                # Copied from `logging.shutdown`.
                try:
                    handler.acquire()
                    handler.flush()
                    handler.close()
                except (OSError, ValueError):
                    pass
                finally:
                    handler.release()
                logger.removeHandler(handler)

    _logger_names = []

    global _log_path
    _log_path = _get_formatted_log_path(log_path, time_stamped)
    return _log_path


def set_logger(
    name: str, sub_path=None, return_logger=False
) -> Optional[logging.Logger]:

    log_path = _get_log_path()

    if sub_path:
        log_path = os.path.join(log_path, sub_path)
        pathlib.Path(log_path).mkdir(parents=True, exist_ok=True)

    global _logger_names
    if name in _logger_names:
        return logging.getLogger(name)

    _logger_names.append(name)

    logger2 = logging.getLogger(name)
    logger2.setLevel(logging.DEBUG)
    ch = _get_handler(log_path, name)
    ch.setLevel(logging.DEBUG)
    # create formatter
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger2.addHandler(ch)
    if return_logger:
        return logger2


def _get_formatted_log_path(log_path: str, time_stamped: bool):
    if log_path is None:
        return
    
    if time_stamped:
        logtime = datetime.timestamp(datetime.now())
        log_path = os.path.join(
            str(log_path),
            "run-" + str(logtime).replace(".", ""),
        )

    pathlib.Path(log_path).mkdir(parents=True, exist_ok=True)
    return log_path


def _get_handler(log_path: str, name: str):
    if log_path:
        # create handler and set level to debug
        log_filename = os.path.join(log_path, name) + ".log"
        return logging.FileHandler(log_filename, mode="w")
    return logging.StreamHandler()



from dataclasses import dataclass
from typing import Optional

from creationism.registration.utils import Text

@dataclass(frozen=True)
class DuplicateRegistrantNameWarning(Warning):
    """Warning for when a registrant name is already registered.

    Args:
        cls (type): The class that contains the register.
        registrant_name (str): The name of the registrant that is already used.
        registrant (type): The registrant to be placed in the register.
        register (dict): The register of cls.
)
    Returns:
        str: Text that explains the warning
    """


    cls: type
    registrant_name: str
    registrant: type
    register: Optional[dict]

    def __str__(self):

        bold = Text.BOLD
        bold_end = Text.BOLD_END

        return (
            f"Unable to register {bold + self.registrant.__name__ + bold_end} with registrant name {bold + self.registrant_name + bold_end} "
            f"because registrant {bold + self.register[self.registrant_name].__name__+ bold_end} is already registerd "
            f"with registrant name {bold + self.registrant_name + bold_end} in the register of {bold + self.cls.__name__ + bold_end}."
        )



import importlib
import operator
import os
import pathlib
import re
from abc import abstractmethod
from collections import UserDict, UserList, UserString
from functools import partial, reduce
from pathlib import Path
from typing import Any

from creationism.configuration.extensions import ConfigurationFileExtension, open_config
from creationism.mode import DEFAULT_MODE
from creationism.registration.factory import RegistrantFactory
from creationism.registration.utils import Text, chain_functions

_REPLACE_IDENTIFIER = "\@replace=(True|true|False|false)"
_REFERENCE_START_SYMBOL = "$"
_REFERENCE_MAP_SYMBOL = ":"
_REFERENCE_ATTRIBUTE_SYMBOL = "."


from copy import deepcopy


class Config(RegistrantFactory):
    CONVERT_NAME = lambda x: chain_functions(
        x, Text.split_capitals, Text.lower, Text.split, partial(Text.get, index=-1)
    )

    @classmethod
    def create(cls, class_type: type, config_value: "Config", *args, **kwargs):
        try:
            if cls.registered(class_type):
                return super().create(
                    class_type, config_value=config_value, *args, **kwargs
                )
            return ConfigObject(config_value, *args, **kwargs)
        except TypeError as te:
            raise te

    @property
    def replace(self):
        return True

    @abstractmethod
    def merge(self, config_value: "Config") -> None:
        """Merges config"""

    @abstractmethod
    def cast(self) -> Any:
        """Casts the data in the config to its original type"""

    @abstractmethod
    def build(self):
        """[summary]"""


def _determine_replace(config1: Config, config2: Config):
    if config2.replace is not None:
        return config2.replace
    elif config1.replace is not None:
        return config1.replace
    return config2.__class__.REPLACE


class UserObject:
    def __init__(self, value):
        self.data = value


class ConfigObject(Config, UserObject):
    def __init__(self, config_value):
        super().__init__(config_value)

    def merge(self, config_value):
        self.data = config_value.data

    def cast(self):
        return self.data

    def build(self, configuration):
        return self

    def __repr__(self):
        return repr(self.data)


@Config.register((str,))
class ConfigString(Config, UserString):
    def __init__(self, config_value):
        super().__init__(config_value)

    def merge(self, config_value):
        self.data = config_value.data

    def cast(self):
        return self.data

    def build(self, configuration):
        if _REFERENCE_START_SYMBOL == self.data[0]:
            return self._get_reference(configuration)
        return self

    def _get_reference(self, configuration):
        self.data = self.data.replace("{", "").replace("}", "")
        references = self.data[1:].split(_REFERENCE_MAP_SYMBOL)
        reference = reduce(operator.getitem, references[:-1], configuration)
        attributes = references[-1].split(_REFERENCE_ATTRIBUTE_SYMBOL)
        reference = reference[attributes[0]].cast()
        for attr in attributes[1:]:
            reference = getattr(reference, attr)
        return ConfigObject(reference)


@Config.register((list,))
class ConfigList(Config, UserList):

    REPLACE = True

    def __init__(self, config_value, replace=None):
        super().__init__(config_value)
        self._replace = replace
        for idx in range(len(self)):
            self.data[idx] = Config.create(
                class_type=type(self[idx]), config_value=self.data[idx]
            )

    @property
    def replace(self):
        return self._replace

    def merge(self, config_value):
        """merge only possible by replace of extend, due to the ambiguity of list items ids"""

        if not isinstance(config_value, ConfigList):
            raise ValueError("unsupporterd merging of different config types")

        if _determine_replace(self, config_value):
            self.data = config_value.data
        else:
            self.data.extend(config_value.data)

    def cast(self):
        return [item.cast() for item in self]

    def build(self, configuration):
        self.data = [item.build(configuration) for item in self.data]
        return self


@Config.register((dict,))
class ConfigDict(Config, UserDict):
    REPLACE = False

    def __init__(self, config_value, replace=None):
        super().__init__(dict())
        self._replace = replace
        self._iterative_init(config_value=config_value)

    @property
    def replace(self):
        return self._replace

    def _iterative_init(self, config_value):
        for key in list(config_value):
            class_type = type(config_value[key])
            replace = re.search(_REPLACE_IDENTIFIER, key)
            if replace:
                key_stripped = key.replace(replace.group(0), "")
                replace = replace.group(1).lower() == "true"
                self.data[key_stripped] = self._create_config(
                    class_type, key, config_value, replace
                )
            else:
                self.data[key] = Config.create(
                    class_type=class_type, config_value=config_value[key]
                )

    def _create_config(self, class_type, key, config_value, replace):
        try:
            return Config.create(
                class_type=class_type, config_value=config_value[key], replace=replace
            )
        except TypeError:
            raise ValueError("create config error")

    def merge(self, config_value):
        if not isinstance(config_value, ConfigDict):
            raise ValueError("unsupporterd merging of different config types")
        else:
            self._iterative_merge(config_value)

    def _iterative_merge(self, config_value):

        if _determine_replace(self, config_value):
            self.data = config_value.data
            return

        for key in list(config_value):
            if key not in self.data:
                self.data[key] = config_value[key]
            else:
                self.data[key].merge(config_value[key])

    def cast(self):
        try:
            return {key: value.cast() for key, value in self.items()}
        except:
            for key, value in self.items():
                print(key, value, type(value))
            raise ValueError()

    def build(self, configuration):
        registrar_module = self.data.pop("registrar_module", None)
        registrar_name = self.data.pop("registrar_name", None)
        registrant_module = self.data.pop("registrant_module", None)
        registrant_name = self.data.pop("registrant_name", None)

        module = self.data.pop("module", None)
        attribute = self.data.pop("attribute", None)

        for key, value in self.items():
            self.data[key] = value.build(configuration)

        if (
            registrar_module is not None
            and registrar_name is not None
            and registrant_name is not None
        ):
            return ConfigObject(
                _build_registrant_object(
                    self,
                    registrar_module,
                    registrar_name,
                    registrant_module,
                    registrant_name,
                )
            )

        if module is not None and attribute is not None:
            return ConfigObject(_build_object(self, module, attribute))

        return self


class Configuration(ConfigDict):

    PRESETS_FOLDER = pathlib.Path(__file__).absolute().parent / "presets"
    CONFIG_PATH = pathlib.Path(__file__).absolute().parent / os.path.join(
        "config", "config.yml"
    )
    SEARCH_PATHS = ("", pathlib.Path(__file__).absolute().parent)

    def __init__(self, name, config_value, modes, search_paths=()):
        self._modes = modes
        self._name = name
        self._search_paths = self.__class__.SEARCH_PATHS + search_paths
        include_configs(config_value, self._search_paths)
        super().__init__(config_value=config_value)

    @classmethod
    def build(
        cls,
        user_config,
        modes,
        build_instances=True,
        build_key=None,
        presets=(),
        external_configurations=None,
        search_paths=(),
        *args,
        **kwargs,
    ):

        if isinstance(user_config, (str, Path)):
            search_paths = cls.SEARCH_PATHS + (Path(user_config).parent,) + search_paths
            user_config = open_config(user_config, search_paths=search_paths)[cls.NAME]
        else:
            search_paths = cls.SEARCH_PATHS + search_paths

        include_configs(user_config, search_paths=search_paths)
        configuration = cls(*args, **kwargs)
        for preset in presets:
            preset_config = open_config(
                cls.PRESETS_FOLDER / (preset + ".yml"), search_paths
            )
            configuration.merge(preset_config)

        configuration.merge(user_config)
        configuration = resolve_modes(configuration, modes=modes)
        resolve_none(configuration)
        if build_instances:
            return configuration.build_instances(external_configurations=external_configurations, build_key=build_key)
        return configuration

    @property
    def name(self):
        return self._name

    def merge(self, config):
        config = ConfigDict(config)
        super().merge(config)

    def build_instances(self, external_configurations=None, build_key=None):
        build = deepcopy(self)

        for mode in self._modes:
            configurations = {self._name: build[mode]}
            if external_configurations is not None:
                for configuration in external_configurations:
                    configurations.update({configuration.name: configuration[mode]})

            build_keys = list(build[mode].keys())
            if build_key is not None:
                index = build_keys.index(build_key) + 1
                build_keys = build_keys[:index]
            
            for key in build_keys:
                build[mode][key] = build[mode][key].build(configurations)

        build = build.cast()
        if build_key:
            return {mode: build[mode][build_key] for mode in self._modes}
        return {self.name: {mode: build[mode] for mode in self._modes}}


def get_module(module):
    try:
        return importlib.import_module(module)
    except:
        pass

    try:
        module = Path(module)
        spec = importlib.util.spec_from_file_location(module.stem, str(module))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except AttributeError as ae:
        raise AttributeError(f"failed to load module: {module}") from ae
    return module


def _build_object(builds, module, attribute):
    module = get_module(module.cast())
    attributes = attribute.cast().split(".")
    attribute = module
    for attr in attributes:
        attribute = getattr(attribute, attr)
    return attribute(**builds.cast())


def _build_registrant_object(
    builds: dict, registrar_module, registrar_name, registrant_module, registrant_name
):
    registrar_module = get_module(registrar_module.cast())
    registrar = getattr(registrar_module, registrar_name.cast())

    # load and register registrant
    if registrant_module is not None:
        _ = get_module(registrant_module.cast())
    builds = builds.cast()
    return registrar.create(registrant_name.cast(), **builds)


def resolve_none(configuration):
    for key, value in configuration.items():
        if issubclass(type(value), UserDict):
            resolve_none(value)
        elif issubclass(type(value), UserString) and value.lower() == "none":
            configuration[key] = ConfigObject(None)


def resolve_modes(config, modes, default_mode=DEFAULT_MODE):
    copied_config = deepcopy(config)
    new_config = Configuration(config.name, dict(), modes=modes)

    if default_mode in copied_config:
        new_config[default_mode] = copied_config[default_mode]
    for mode in modes:
        new_config[mode] = deepcopy(new_config[default_mode])
        if mode in copied_config:
            new_config[mode].merge(copied_config[mode])
    return new_config


def include_configs(config, search_paths=()):
    search_paths = ("",) + search_paths

    for k, v in config.items():
        if isinstance(v, dict):
            include_configs(v, search_paths)

        elif isinstance(
            v, str
        ) and ConfigurationFileExtension.is_configuration_extension_path(v):
            config[k] = open_config(v, search_paths)



from functools import lru_cache
from typing import Union

from creationism.registration.registrar import Registrar

cache = lru_cache(maxsize=None)
from dataclasses import dataclass
from typing import Optional
import inspect

@dataclass(frozen=True)
class RegistrantNotRegisteredError(Exception):
    """Raised when registrant name does not exist in the register"""

    cls: type
    registrant_name: str
    register: Optional[dict]

    def __post_init__(self):
        super().__init__(self._message())

    def _message(self):
        if self.register is None:
            return self._empty_register_message()
        return self._register_name_not_found_message()

    def _prefix_message(self):
        return f"Registrant '{self.registrant_name}' is not found in the register of class '{self.cls.__name__}'"

    def _empty_register_message(self):
        return f"""
        {self._prefix_message()} with an empty register.
        """

    def _register_name_not_found_message(self):
        return f"""
        {self._prefix_message()} with registrant names {tuple(self.register.keys())} and register classes {tuple(self.register.values())}.
        """



class RegistrantFactory(Registrar):
    """Factory class for creating instances or retreive object references

    Class Atributes:
        STATIC (bool): Used to create static instances if the instantiation is identical

    Raises:
        RegistrantNotRegisteredError: raised when registrant name is not in the register

    """

    STATIC = False

    @classmethod
    def create(
        cls,
        registrant_name: Union[str, type],
        return_type: bool = False,
        *args,
        **kwargs
    ):
        if cls._REGISTER is None:
            raise RegistrantNotRegisteredError(
                cls=cls,
                registrant_name=registrant_name,
                register=cls._REGISTER,
            )

        if issubclass(type(registrant_name), cls):
            return registrant_name

        if callable(registrant_name) and registrant_name in cls._REGISTER.values():
            class_type = registrant_name
        elif registrant_name in cls._REGISTER:
            class_type = cls.get_registrant(registrant_name)
        else:
            raise RegistrantNotRegisteredError(
                cls=cls,
                registrant_name=registrant_name,
                register=cls._REGISTER,
            )

        if return_type:
            return class_type

        if cls.STATIC:
            return cls._static_create(class_type, *args, **kwargs)
        return cls._create(class_type, *args, **kwargs)

    @classmethod
    @cache
    def _static_create(cls, class_type, *args, **kwargs):
        return cls._create(class_type, *args, **kwargs)

    @classmethod
    def _create(cls, class_type, *args, **kwargs):
        return class_type(*args, **kwargs)



from setuptools import setup, find_packages

setup(
    name='creationism',
    version='0.0.3',
    author='Mart van Rijthoven',
    author_email='mart.vanrijthoven@gmail.com',
    packages=find_packages(exclude=("tests",'notebooks')),
    url='http://pypi.python.org/pypi/creationism/',
    license='LICENSE.txt',
    long_description='Creationism enables you to create like a God',
)


from abc import abstractstaticmethod
from pathlib import Path
from creationism.registration.factory import RegistrantFactory

class Extension(RegistrantFactory):
    STATIC = True

    @classmethod
    def has_extension(cls, path: Path, extension: "Extension"):
        return cls.create(path.suffix, return_type=True) == extension
    
    @classmethod
    def is_extension(cls, extension_name, extension):
        return issubclass(cls.create(extension_name, return_type=True), extension)





import json

import yaml
from functools import partial
import re
from creationism.registration.utils import Text, chain_functions

def open_json(file_path):
    with open(str(file_path)) as file:
        return json.load(file)


def open_yaml(file_path):
    with open(str(file_path)) as file:
        return yaml.load(file, Loader=yaml.Loader)


def first_lowered(x):
    return chain_functions(
        x, Text.split_capitals, Text.lower, Text.split, partial(Text.get, index=0)
    )
def create_name_from_class_name(name):
    return " ".join(re.findall(".[^A-Z]*", name)[:-1])



import re
from typing import Any, Callable, Tuple


class Text:
    BOLD = "\033[1m"
    BOLD_END = "\033[0m"

    @staticmethod
    def lower(text: str) -> str:
        return text.lower()

    @staticmethod
    def split_capitals(text: str, sep: str = ' ') -> str:
        return sep.join(re.findall(".[^A-Z]*", text))

    @staticmethod
    def split(text: str, sep: str = ' ') -> str:
        return text.split(sep)

    @staticmethod
    def get(texts: str, index: int = 0) -> str:
        return texts[index]


def chain_functions(x: Any, *functions: Tuple[Callable]) -> Any:
    for function in functions:
        x = function(x)
    return x
